@@ -4,6 +4,26 @@ from __future__ import division
 
 __copyright__ = "Copyright (C) 2011 Andreas Kloeckner"
 
+__license__ = """
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in
+all copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+THE SOFTWARE.
+"""
+
 import pyopencl._cl as _cl
 import re
 import sys
@@ -19,8 +39,6 @@ except ImportError:
     new_hash = md5.new
 
 
-
-
 def _erase_dir(dir):
     from os import listdir, unlink, rmdir
     from os.path import join
@@ -28,6 +46,7 @@ def _erase_dir(dir):
         unlink(join(dir, name))
     rmdir(dir)
 
+
 def update_checksum(checksum, obj):
     if isinstance(obj, unicode):
         checksum.update(obj.encode("utf8"))
@@ -35,14 +54,12 @@ def update_checksum(checksum, obj):
         checksum.update(obj)
 
 
-
-
-
 # {{{ cleanup
 
 class CleanupBase(object):
     pass
 
+
 class CleanupManager(CleanupBase):
     def __init__(self):
         self.cleanups = []
@@ -58,6 +75,7 @@ class CleanupManager(CleanupBase):
         for c in self.cleanups:
             c.error_clean_up()
 
+
 class CacheLockManager(CleanupBase):
     def __init__(self, cleanup_m, cache_dir):
         if cache_dir is not None:
@@ -92,6 +110,7 @@ class CacheLockManager(CleanupBase):
     def error_clean_up(self):
         pass
 
+
 class ModuleCacheDirManager(CleanupBase):
     def __init__(self, cleanup_m, path):
         from os import mkdir
@@ -121,11 +140,13 @@ class ModuleCacheDirManager(CleanupBase):
 
 # }}}
 
+
 # {{{ #include dependency handling
 
 C_INCLUDE_RE = re.compile(r'^\s*\#\s*include\s+[<"](.+)[">]\s*$',
         re.MULTILINE)
 
+
 def get_dependencies(src, include_path):
     result = {}
 
@@ -150,7 +171,8 @@ def get_dependencies(src, include_path):
                     finally:
                         src_file.close()
 
-                    # prevent infinite recursion if some header file appears to include itself
+                    # jrevent infinite recursion if some header file appears to
+                    # include itself
                     result[included_file_name] = None
 
                     checksum = new_hash()
@@ -163,7 +185,7 @@ def get_dependencies(src, include_path):
                             )
 
                     found = True
-                    break # stop searching the include path
+                    break  # stop searching the include path
 
             if not found:
                 pass
@@ -176,8 +198,6 @@ def get_dependencies(src, include_path):
     return result
 
 
-
-
 def get_file_md5sum(fname):
     checksum = new_hash()
     inf = open(fname)
@@ -189,13 +209,11 @@ def get_file_md5sum(fname):
     return checksum.hexdigest()
 
 
-
-
 def check_dependencies(deps):
     for name, date, md5sum in deps:
         try:
             possibly_updated = os.stat(name).st_mtime != date
-        except OSError, e:
+        except OSError:
             return False
         else:
             if possibly_updated and md5sum != get_file_md5sum(name):
@@ -205,18 +223,17 @@ def check_dependencies(deps):
 
 # }}}
 
+
 # {{{ key generation
 
 def get_device_cache_id(device):
     from pyopencl.version import VERSION
     platform = device.platform
-    return (VERSION, 
+    return (VERSION,
             platform.vendor, platform.name, platform.version,
             device.vendor, device.name, device.version, device.driver_version)
 
 
-
-
 def get_cache_key(device, options, src):
     checksum = new_hash()
     update_checksum(checksum, src)
@@ -227,9 +244,6 @@ def get_cache_key(device, options, src):
 # }}}
 
 
-
-
-
 def retrieve_from_cache(cache_dir, cache_key):
     class _InvalidInfoFile(RuntimeError):
         pass
@@ -241,66 +255,71 @@ def retrieve_from_cache(cache_dir, cache_key):
 
     cleanup_m = CleanupManager()
     try:
-        lock_m = CacheLockManager(cleanup_m, cache_dir)
-
-        mod_cache_dir_m = ModuleCacheDirManager(cleanup_m, module_cache_dir)
-        info_path = mod_cache_dir_m.sub("info")
-        binary_path = mod_cache_dir_m.sub("binary")
-        source_path = mod_cache_dir_m.sub("source.cl")
-
-        # {{{ load info file
-
         try:
-            from cPickle import load
+            CacheLockManager(cleanup_m, cache_dir)
+
+            mod_cache_dir_m = ModuleCacheDirManager(cleanup_m, module_cache_dir)
+            info_path = mod_cache_dir_m.sub("info")
+            binary_path = mod_cache_dir_m.sub("binary")
+
+            # {{{ load info file
 
             try:
-                info_file = open(info_path, "rb")
-            except IOError:
-                raise _InvalidInfoFile()
+                from cPickle import load
 
+                try:
+                    info_file = open(info_path, "rb")
+                except IOError:
+                    raise _InvalidInfoFile()
+
+                try:
+                    try:
+                        info = load(info_file)
+                    except EOFError:
+                        raise _InvalidInfoFile()
+                finally:
+                    info_file.close()
+
+            except _InvalidInfoFile:
+                mod_cache_dir_m.reset()
+                from warnings import warn
+                warn("PyOpenCL encountered an invalid info file for cache key %s"
+                        % cache_key)
+                return None
+
+            # }}}
+
+            # {{{ load binary
+
+            binary_file = open(binary_path, "rb")
             try:
-                info = load(info_file)
-            except EOFError:
-                raise _InvalidInfoFile()
+                binary = binary_file.read()
             finally:
-                info_file.close()
+                binary_file.close()
 
-        except _InvalidInfoFile:
-            mod_cache_dir_m.reset()
-            from warnings import warn
-            warn("PyOpenCL encountered an invalid info file for cache key %s" 
-                    % cache_key)
-            return None
+            # }}}
 
-        # }}}
+            if check_dependencies(info.dependencies):
+                return binary, info.log
+            else:
+                mod_cache_dir_m.reset()
 
-        # {{{ load binary
-
-        binary_file = open(binary_path, "rb")
-        try:
-            binary = binary_file.read()
-        finally:
-            binary_file.close()
-
-        # }}}
-
-        if check_dependencies(info.dependencies):
-            return binary, info.log
-        else:
-            mod_cache_dir_m.reset()
-
-    except:
-        cleanup_m.error_clean_up()
-        raise
+        except:
+            cleanup_m.error_clean_up()
+            raise
     finally:
         cleanup_m.clean_up()
 
+
 # {{{ top-level driver
 
 class _SourceInfo(Record):
     pass
 
+
 def _create_built_program_from_source_cached(ctx, src, options, devices, cache_dir):
+    from os.path import join
+
     include_path = ["."]
 
     option_idx = 0
@@ -312,13 +331,12 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
                     include_path.append(options[option_idx+1])
                 option_idx += 2
             else:
-                include_path.append(option[2:])
+                include_path.append(option[2:].lstrip())
                 option_idx += 1
         else:
             option_idx += 1
 
     if cache_dir is None:
-        from os.path import join
         from tempfile import gettempdir
         import getpass
         cache_dir = join(gettempdir(),
@@ -385,10 +403,11 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
         prg_bins = prg.get_info(_cl.program_info.BINARIES)
         prg_logs = prg._get_build_logs()
 
-        for i, dest_index in enumerate(to_be_built_indices):
-            assert prg_devs[i] == devices[dest_index]
-            binaries[dest_index] = prg_bins[i]
-            _, logs[dest_index] = prg_logs[i]
+        for dest_index in to_be_built_indices:
+            dev = devices[dest_index]
+            src_index = prg_devs.index(dev)
+            binaries[dest_index] = prg_bins[src_index]
+            _, logs[dest_index] = prg_logs[src_index]
 
         if len(to_be_built_indices) == len(devices):
             # Important special case: if code for all devices was built,
@@ -408,37 +427,38 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
     if to_be_built_indices:
         cleanup_m = CleanupManager()
         try:
-            lock_m = CacheLockManager(cleanup_m, cache_dir)
+            try:
+                CacheLockManager(cleanup_m, cache_dir)
 
-            for i in to_be_built_indices:
-                cache_key = cache_keys[i]
-                device = devices[i]
-                binary = binaries[i]
+                for i in to_be_built_indices:
+                    cache_key = cache_keys[i]
+                    device = devices[i]
+                    binary = binaries[i]
 
-                mod_cache_dir_m = ModuleCacheDirManager(cleanup_m,
-                        join(cache_dir, cache_key))
-                info_path = mod_cache_dir_m.sub("info")
-                binary_path = mod_cache_dir_m.sub("binary")
-                source_path = mod_cache_dir_m.sub("source.cl")
+                    mod_cache_dir_m = ModuleCacheDirManager(cleanup_m,
+                            join(cache_dir, cache_key))
+                    info_path = mod_cache_dir_m.sub("info")
+                    binary_path = mod_cache_dir_m.sub("binary")
+                    source_path = mod_cache_dir_m.sub("source.cl")
 
-                outf = open(source_path, "wt")
-                outf.write(src)
-                outf.close()
+                    outf = open(source_path, "wt")
+                    outf.write(src)
+                    outf.close()
 
-                outf = open(binary_path, "wb")
-                outf.write(binary)
-                outf.close()
+                    outf = open(binary_path, "wb")
+                    outf.write(binary)
+                    outf.close()
 
-                from cPickle import dump
-                info_file = open(info_path, "wb")
-                dump(_SourceInfo(
-                    dependencies=get_dependencies(src, include_path),
-                    log=logs[i]), info_file)
-                info_file.close()
+                    from cPickle import dump
+                    info_file = open(info_path, "wb")
+                    dump(_SourceInfo(
+                        dependencies=get_dependencies(src, include_path),
+                        log=logs[i]), info_file)
+                    info_file.close()
 
-        except:
-            cleanup_m.error_clean_up()
-            raise
+            except:
+                cleanup_m.error_clean_up()
+                raise
         finally:
             cleanup_m.clean_up()
 
@@ -447,12 +467,10 @@ def _create_built_program_from_source_cached(ctx, src, options, devices, cache_d
     return result, already_built
 
 
-
-
 def create_built_program_from_source_cached(ctx, src, options=[], devices=None,
         cache_dir=None):
     try:
-        if cache_dir != False:
+        if cache_dir is not False:
             prg, already_built = _create_built_program_from_source_cached(
                     ctx, src, options, devices, cache_dir)
         else:
@@ -462,7 +480,7 @@ def create_built_program_from_source_cached(ctx, src, options=[], devices=None,
     except Exception, e:
         raise
         from pyopencl import Error
-        if (isinstance(e, Error) 
+        if (isinstance(e, Error)
                 and e.code == _cl.status_code.BUILD_PROGRAM_FAILURE):
             # no need to try again
             raise
