@@ -31,30 +31,28 @@ None of the original source code remains.
 """
 
 
-
-
 import pyopencl as cl
 from pyopencl.tools import (
         context_dependent_memoize,
-        dtype_to_ctype)
+        dtype_to_ctype, KernelTemplateBase,
+        _process_code_for_macro)
 import numpy as np
-import pyopencl._mymako as mako
 
 
+# {{{ kernel source
 
-
-KERNEL = """
-
+KERNEL = """//CL//
     #define GROUP_SIZE ${group_size}
     #define READ_AND_MAP(i) (${map_expr})
     #define REDUCE(a, b) (${reduce_expr})
 
     % if double_support:
         #pragma OPENCL EXTENSION cl_khr_fp64: enable
-    % elif amd_double_support:
-        #pragma OPENCL EXTENSION cl_amd_fp64: enable
+        #define PYOPENCL_DEFINE_CDOUBLE
     % endif
 
+    #include <pyopencl-complex.h>
+
     ${preamble}
 
     typedef ${out_type} out_type;
@@ -63,6 +61,8 @@ KERNEL = """
       __global out_type *out, ${arguments},
       unsigned int seq_count, unsigned int n)
     {
+       ${arg_prep}
+
         __local out_type ldata[GROUP_SIZE];
 
         unsigned int lid = get_local_id(0);
@@ -110,7 +110,15 @@ KERNEL = """
 
             barrier(CLK_LOCAL_MEM_FENCE);
 
-            if (lid < ${no_sync_size})
+            <%
+            # NB: There's an exact duplicate of this calculation in the
+            # %while loop below.
+
+            new_size = cur_size // 2
+            assert new_size * 2 == cur_size
+            %>
+
+            if (lid < ${new_size})
             {
                 __local volatile out_type *lvdata = ldata;
                 % while cur_size > 1:
@@ -134,14 +142,16 @@ KERNEL = """
     }
     """
 
+# }}}
 
 
+# {{{ internal codegen frontends
 
-def  get_reduction_source(
-         ctx, out_type, out_type_size,
-         neutral, reduce_expr, map_expr, arguments,
-         name="reduce_kernel", preamble="",
-         device=None, max_group_size=None):
+def _get_reduction_source(
+        ctx, out_type, out_type_size,
+        neutral, reduce_expr, map_expr, parsed_args,
+        name="reduce_kernel", preamble="", arg_prep="",
+        device=None, max_group_size=None):
 
     if device is not None:
         devices = [device]
@@ -155,10 +165,15 @@ def  get_reduction_source(
         max_work_group_size = device.max_work_group_size
         if "RV770" in device.name:
             max_work_group_size = 64
-        return min(
-                max_work_group_size,
-                (device.local_mem_size + out_type_size - 1)
-                // out_type_size)
+
+        # compute lmem limit
+        from pytools import div_ceil
+        lmem_wg_size = div_ceil(max_work_group_size, out_type_size)
+        result = min(max_work_group_size, lmem_wg_size)
+
+        # round down to power of 2
+        from pyopencl.tools import bitlog2
+        return 2**bitlog2(result)
 
     group_size = min(get_dev_group_size(dev) for dev in devices)
 
@@ -187,24 +202,23 @@ def  get_reduction_source(
 
     from mako.template import Template
     from pytools import all
-    from pyopencl.characterize import has_double_support, has_amd_double_support
+    from pyopencl.characterize import has_double_support
     src = str(Template(KERNEL).render(
         out_type=out_type,
-        arguments=arguments,
+        arguments=", ".join(arg.declarator() for arg in parsed_args),
         group_size=group_size,
         no_sync_size=no_sync_size,
         neutral=neutral,
-        reduce_expr=reduce_expr,
-        map_expr=map_expr,
+        reduce_expr=_process_code_for_macro(reduce_expr),
+        map_expr=_process_code_for_macro(map_expr),
         name=name,
         preamble=preamble,
-        double_support=all(
-            has_double_support(dev) for dev in devices),
-        amd_double_support=all(
-            has_amd_double_support(dev) for dev in devices)
+        arg_prep=arg_prep,
+        double_support=all(has_double_support(dev) for dev in devices),
         ))
 
     from pytools import Record
+
     class ReductionInfo(Record):
         pass
 
@@ -214,52 +228,55 @@ def  get_reduction_source(
             group_size=group_size)
 
 
-
-
 def get_reduction_kernel(stage,
-         ctx, out_type, out_type_size,
+         ctx, dtype_out,
          neutral, reduce_expr, map_expr=None, arguments=None,
          name="reduce_kernel", preamble="",
          device=None, options=[], max_group_size=None):
+
     if map_expr is None:
         if stage == 2:
             map_expr = "pyopencl_reduction_inp[i]"
         else:
             map_expr = "in[i]"
 
-    if stage == 2:
-        in_arg = "__global const %s *pyopencl_reduction_inp" % out_type
-        if arguments:
-            arguments = in_arg + ", " + arguments
-        else:
-            arguments = in_arg
+    from pyopencl.tools import (
+            parse_arg_list, get_arg_list_scalar_arg_dtypes,
+            get_arg_offset_adjuster_code, VectorArg)
 
-    inf = get_reduction_source(
-            ctx, out_type, out_type_size,
+    arg_prep = ""
+    if stage == 1 and arguments is not None:
+        arguments = parse_arg_list(arguments, with_offset=True)
+        arg_prep = get_arg_offset_adjuster_code(arguments)
+
+    if stage == 2 and arguments is not None:
+        arguments = parse_arg_list(arguments)
+        arguments = (
+                [VectorArg(dtype_out, "pyopencl_reduction_inp")]
+                + arguments)
+
+    inf = _get_reduction_source(
+            ctx, dtype_to_ctype(dtype_out), dtype_out.itemsize,
             neutral, reduce_expr, map_expr, arguments,
-            name, preamble, device, max_group_size)
+            name, preamble, arg_prep, device, max_group_size)
 
     inf.program = cl.Program(ctx, inf.source)
     inf.program.build(options)
     inf.kernel = getattr(inf.program, name)
 
-    from pyopencl.tools import parse_c_arg, ScalarArg
+    inf.arg_types = arguments
 
-    inf.arg_types = [parse_c_arg(arg) for arg in arguments.split(",")]
-    scalar_arg_dtypes = [None]
-    for arg_type in inf.arg_types:
-        if isinstance(arg_type, ScalarArg):
-            scalar_arg_dtypes.append(arg_type.dtype)
-        else:
-            scalar_arg_dtypes.append(None)
-    scalar_arg_dtypes.extend([np.uint32]*2)
-
-    inf.kernel.set_scalar_arg_dtypes(scalar_arg_dtypes)
+    inf.kernel.set_scalar_arg_dtypes(
+            [None]
+            + get_arg_list_scalar_arg_dtypes(inf.arg_types)
+            + [np.uint32]*2)
 
     return inf
 
+# }}}
 
 
+# {{{ main reduction kernel
 
 class ReductionKernel:
     def __init__(self, ctx, dtype_out,
@@ -273,7 +290,7 @@ class ReductionKernel:
 
         while True:
             self.stage_1_inf = get_reduction_kernel(1, ctx,
-                    dtype_to_ctype(dtype_out), dtype_out.itemsize,
+                    dtype_out,
                     neutral, reduce_expr, map_expr, arguments,
                     name=name+"_stage1", options=options, preamble=preamble,
                     max_group_size=max_group_size)
@@ -291,7 +308,7 @@ class ReductionKernel:
             assert trip_count <= 2
 
         self.stage_2_inf = get_reduction_kernel(2, ctx,
-                dtype_to_ctype(dtype_out), dtype_out.itemsize,
+                dtype_out,
                 neutral, reduce_expr, arguments=arguments,
                 name=name+"_stage2", options=options, preamble=preamble,
                 max_group_size=max_group_size)
@@ -301,8 +318,8 @@ class ReductionKernel:
         assert any(
                 isinstance(arg_tp, VectorArg)
                 for arg_tp in self.stage_1_inf.arg_types), \
-                "ReductionKernel can only be used with functions that have at least one " \
-                "vector argument"
+                "ReductionKernel can only be used with functions " \
+                "that have at least one vector argument"
 
     def __call__(self, *args, **kwargs):
         MAX_GROUP_COUNT = 1024
@@ -313,6 +330,8 @@ class ReductionKernel:
         stage_inf = self.stage_1_inf
 
         queue = kwargs.pop("queue", None)
+        wait_for = kwargs.pop("wait_for", None)
+        return_event = kwargs.pop("return_event", False)
 
         if kwargs:
             raise TypeError("invalid keyword argument to reduction kernel")
@@ -331,7 +350,9 @@ class ReductionKernel:
                                 "deal with non-contiguous arrays")
 
                     vectors.append(arg)
-                    invocation_args.append(arg.data)
+                    invocation_args.append(arg.base_data)
+                    if arg_tp.with_offset:
+                        invocation_args.append(arg.offset)
                 else:
                     invocation_args.append(arg)
 
@@ -361,20 +382,85 @@ class ReductionKernel:
                         (group_count,), self.dtype_out,
                         allocator=repr_vec.allocator)
 
-            stage_inf.kernel(
+            last_evt = stage_inf.kernel(
                     use_queue,
                     (group_count*stage_inf.group_size,),
                     (stage_inf.group_size,),
-                    *([result.data]+invocation_args+[seq_count, sz]))
+                    *([result.data]+invocation_args+[seq_count, sz]),
+                    **dict(wait_for=wait_for))
+            wait_for = [last_evt]
 
             if group_count == 1:
-                return result
+                if return_event:
+                    return result, last_evt
+                else:
+                    return result
             else:
                 stage_inf = self.stage_2_inf
                 args = (result,) + stage1_args
 
+# }}}
 
 
+# {{{ template
+
+class ReductionTemplate(KernelTemplateBase):
+    def __init__(self,
+            arguments, neutral, reduce_expr, map_expr=None,
+            is_segment_start_expr=None, input_fetch_exprs=[],
+            name_prefix="reduce", preamble="", template_processor=None):
+
+        KernelTemplateBase.__init__(
+                self, template_processor=template_processor)
+        self.arguments = arguments
+        self.reduce_expr = reduce_expr
+        self.neutral = neutral
+        self.map_expr = map_expr
+        self.name_prefix = name_prefix
+        self.preamble = preamble
+
+    def build_inner(self, context, type_aliases=(), var_values=(),
+            more_preamble="", more_arguments=(), declare_types=(),
+            options=(), devices=None):
+        renderer = self.get_renderer(
+                type_aliases, var_values, context, options)
+
+        arg_list = renderer.render_argument_list(
+                self.arguments, more_arguments)
+
+        type_decl_preamble = renderer.get_type_decl_preamble(
+                context.devices[0], declare_types, arg_list)
+
+        return ReductionKernel(context, renderer.type_aliases["reduction_t"],
+                renderer(self.neutral), renderer(self.reduce_expr),
+                renderer(self.map_expr),
+                renderer.render_argument_list(self.arguments, more_arguments),
+                name=renderer(self.name_prefix), options=list(options),
+                preamble=(
+                    type_decl_preamble
+                    + "\n"
+                    + renderer(self.preamble + "\n" + more_preamble)))
+
+# }}}
+
+
+# {{{ array reduction kernel getters
+
+@context_dependent_memoize
+def get_any_kernel(ctx, dtype_in):
+    from pyopencl.tools import VectorArg
+    return ReductionKernel(ctx, np.int8, "false", "a || b",
+            map_expr="(bool) (in[i])",
+            arguments=[VectorArg(dtype_in, "in")])
+
+
+@context_dependent_memoize
+def get_all_kernel(ctx, dtype_in):
+    from pyopencl.tools import VectorArg
+    return ReductionKernel(ctx, np.int8, "true", "a && b",
+            map_expr="(bool) (in[i])",
+            arguments=[VectorArg(dtype_in, "in")])
+
 
 @context_dependent_memoize
 def get_sum_kernel(ctx, dtype_out, dtype_in):
@@ -382,67 +468,112 @@ def get_sum_kernel(ctx, dtype_out, dtype_in):
         dtype_out = dtype_in
 
     return ReductionKernel(ctx, dtype_out, "0", "a+b",
-            arguments="__global const %(tp)s *in"
+            arguments="const %(tp)s *in"
             % {"tp": dtype_to_ctype(dtype_in)})
 
 
-
-
-@context_dependent_memoize
-def get_dot_kernel(ctx, dtype_out, dtype_a=None, dtype_b=None):
-    if dtype_out is None:
-        dtype_out = dtype_a
-
+def _get_dot_expr(dtype_out, dtype_a, dtype_b, conjugate_first,
+        has_double_support, index_expr="i"):
     if dtype_b is None:
         if dtype_a is None:
             dtype_b = dtype_out
         else:
             dtype_b = dtype_a
 
-    if dtype_a is None:
-        dtype_a = dtype_out
+    if dtype_out is None:
+        from pyopencl.compyte.array import get_common_dtype
+        dtype_out = get_common_dtype(
+                dtype_a.type(0), dtype_b.type(0),
+                has_double_support)
+
+    a_real_dtype = dtype_a.type(0).real.dtype
+    b_real_dtype = dtype_b.type(0).real.dtype
+    out_real_dtype = dtype_out.type(0).real.dtype
+
+    a_is_complex = dtype_a.kind == "c"
+    b_is_complex = dtype_b.kind == "c"
+    out_is_complex = dtype_out.kind == "c"
+
+    from pyopencl.elementwise import complex_dtype_to_name
+
+    if a_is_complex and b_is_complex:
+        a = "a[%s]" % index_expr
+        b = "b[%s]" % index_expr
+        if dtype_a != dtype_out:
+            a = "%s_cast(%s)" % (complex_dtype_to_name(dtype_out), a)
+        if dtype_b != dtype_out:
+            b = "%s_cast(%s)" % (complex_dtype_to_name(dtype_out), b)
+
+        if conjugate_first and a_is_complex:
+            a = "%s_conj(%s)" % (
+                    complex_dtype_to_name(dtype_out), a)
+
+        map_expr = "%s_mul(%s, %s)" % (
+                complex_dtype_to_name(dtype_out), a, b)
+    else:
+        a = "a[%s]" % index_expr
+        b = "b[%s]" % index_expr
+
+        if out_is_complex:
+            if a_is_complex and dtype_a != dtype_out:
+                a = "%s_cast(%s)" % (complex_dtype_to_name(dtype_out), a)
+            if b_is_complex and dtype_b != dtype_out:
+                b = "%s_cast(%s)" % (complex_dtype_to_name(dtype_out), b)
+
+            if not a_is_complex and a_real_dtype != out_real_dtype:
+                a = "(%s) (%s)" % (dtype_to_ctype(out_real_dtype), a)
+            if not b_is_complex and b_real_dtype != out_real_dtype:
+                b = "(%s) (%s)" % (dtype_to_ctype(out_real_dtype), b)
+
+        if conjugate_first and a_is_complex:
+            a = "%s_conj(%s)" % (
+                    complex_dtype_to_name(dtype_out), a)
+
+        map_expr = "%s*%s" % (a, b)
+
+    return map_expr, dtype_out, dtype_b
+
+
+@context_dependent_memoize
+def get_dot_kernel(ctx, dtype_out, dtype_a=None, dtype_b=None,
+        conjugate_first=False):
+    from pyopencl.characterize import has_double_support
+    map_expr, dtype_out, dtype_b = _get_dot_expr(
+            dtype_out, dtype_a, dtype_b, conjugate_first,
+            has_double_support=has_double_support(ctx.devices[0]))
 
     return ReductionKernel(ctx, dtype_out, neutral="0",
-            reduce_expr="a+b", map_expr="a[i]*b[i]",
+            reduce_expr="a+b", map_expr=map_expr,
             arguments=
-            "__global const %(tp_a)s *a, "
-            "__global const %(tp_b)s *b" % {
+            "const %(tp_a)s *a, "
+            "const %(tp_b)s *b" % {
                 "tp_a": dtype_to_ctype(dtype_a),
                 "tp_b": dtype_to_ctype(dtype_b),
                 })
 
 
-
-
 @context_dependent_memoize
-def get_subset_dot_kernel(ctx, dtype_out, dtype_subset, dtype_a=None, dtype_b=None):
-    if dtype_out is None:
-        dtype_out = dtype_a
-
-    if dtype_b is None:
-        if dtype_a is None:
-            dtype_b = dtype_out
-        else:
-            dtype_b = dtype_a
-
-    if dtype_a is None:
-        dtype_a = dtype_out
+def get_subset_dot_kernel(ctx, dtype_out, dtype_subset, dtype_a=None, dtype_b=None,
+        conjugate_first=False):
+    from pyopencl.characterize import has_double_support
+    map_expr, dtype_out, dtype_b = _get_dot_expr(
+            dtype_out, dtype_a, dtype_b, conjugate_first,
+            has_double_support=has_double_support(ctx.devices[0]),
+            index_expr="lookup_tbl[i]")
 
     # important: lookup_tbl must be first--it controls the length
     return ReductionKernel(ctx, dtype_out, neutral="0",
-            reduce_expr="a+b", map_expr="a[lookup_tbl[i]]*b[lookup_tbl[i]]",
+            reduce_expr="a+b", map_expr=map_expr,
             arguments=
-            "__global const %(tp_lut)s *lookup_tbl, "
-            "__global const %(tp_a)s *a, "
-            "__global const %(tp_b)s *b" % {
+            "const %(tp_lut)s *lookup_tbl, "
+            "const %(tp_a)s *a, "
+            "const %(tp_b)s *b" % {
             "tp_lut": dtype_to_ctype(dtype_subset),
             "tp_a": dtype_to_ctype(dtype_a),
             "tp_b": dtype_to_ctype(dtype_b),
             })
 
 
-
-
 def get_minmax_neutral(what, dtype):
     dtype = np.dtype(dtype)
     if issubclass(dtype.type, np.inexact):
@@ -461,8 +592,6 @@ def get_minmax_neutral(what, dtype):
             raise ValueError("what is not min or max.")
 
 
-
-
 @context_dependent_memoize
 def get_minmax_kernel(ctx, what, dtype):
     if dtype.kind == "f":
@@ -475,13 +604,11 @@ def get_minmax_kernel(ctx, what, dtype):
     return ReductionKernel(ctx, dtype,
             neutral=get_minmax_neutral(what, dtype),
             reduce_expr="%(reduce_expr)s" % {"reduce_expr": reduce_expr},
-            arguments="__global const %(tp)s *in" % {
+            arguments="const %(tp)s *in" % {
                 "tp": dtype_to_ctype(dtype),
                 }, preamble="#define MY_INFINITY (1./0)")
 
 
-
-
 @context_dependent_memoize
 def get_subset_minmax_kernel(ctx, what, dtype, dtype_subset):
     if dtype.kind == "f":
@@ -496,8 +623,12 @@ def get_subset_minmax_kernel(ctx, what, dtype, dtype_subset):
             reduce_expr="%(reduce_expr)s" % {"reduce_expr": reduce_expr},
             map_expr="in[lookup_tbl[i]]",
             arguments=
-            "__global const %(tp_lut)s *lookup_tbl, "
-            "__global const %(tp)s *in"  % {
+            "const %(tp_lut)s *lookup_tbl, "
+            "const %(tp)s *in" % {
             "tp": dtype_to_ctype(dtype),
             "tp_lut": dtype_to_ctype(dtype_subset),
             }, preamble="#define MY_INFINITY (1./0)")
+
+# }}}
+
+# vim: filetype=pyopencl:fdm=marker
