@@ -1,4 +1,5 @@
 #! /usr/bin/env python
+from __future__ import division, with_statement
 
 __copyright__ = "Copyright (C) 2009 Andreas Kloeckner"
 
@@ -25,7 +26,6 @@ THE SOFTWARE.
 import numpy as np
 import numpy.linalg as la
 import sys
-import pytools.test
 
 import pyopencl as cl
 import pyopencl.array as cl_array
@@ -71,7 +71,6 @@ def make_random_array(queue, dtype, size):
 
 # {{{ dtype-related
 
-@pytools.test.mark_test.opencl
 def test_basic_complex(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -88,7 +87,6 @@ def test_basic_complex(ctx_factory):
     assert la.norm((ary*c).get() - c*host_ary) < 1e-5 * la.norm(host_ary)
 
 
-@pytools.test.mark_test.opencl
 def test_mix_complex(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -159,7 +157,6 @@ def test_mix_complex(ctx_factory):
                     assert correct
 
 
-@pytools.test.mark_test.opencl
 def test_pow_neg1_vs_inv(ctx_factory):
     ctx = ctx_factory()
     queue = cl.CommandQueue(ctx)
@@ -179,7 +176,6 @@ def test_pow_neg1_vs_inv(ctx_factory):
     assert la.norm(res2-ref, np.inf) / la.norm(ref) < 1e-13
 
 
-@pytools.test.mark_test.opencl
 def test_vector_fill(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -192,7 +188,6 @@ def test_vector_fill(ctx_factory):
     a_gpu = cl_array.zeros(queue, 100, dtype=cl_array.vec.float4)
 
 
-@pytools.test.mark_test.opencl
 def test_absrealimag(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -227,7 +222,6 @@ def test_absrealimag(ctx_factory):
 
 # {{{ operators
 
-@pytools.test.mark_test.opencl
 def test_rmul_yields_right_type(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -242,7 +236,6 @@ def test_rmul_yields_right_type(ctx_factory):
     assert isinstance(two_a, cl_array.Array)
 
 
-@pytools.test.mark_test.opencl
 def test_pow_array(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -257,7 +250,6 @@ def test_pow_array(ctx_factory):
     assert (np.abs(pow(a, a) - result) < 1e-3).all()
 
 
-@pytools.test.mark_test.opencl
 def test_pow_number(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -269,7 +261,6 @@ def test_pow_number(ctx_factory):
     assert (np.abs(a ** 2 - result) < 1e-3).all()
 
 
-@pytools.test.mark_test.opencl
 def test_multiply(ctx_factory):
     """Test the muliplication of an array with a scalar. """
 
@@ -289,7 +280,6 @@ def test_multiply(ctx_factory):
                 assert (a * scalar == a_mult).all()
 
 
-@pytools.test.mark_test.opencl
 def test_multiply_array(ctx_factory):
     """Test the multiplication of two arrays."""
 
@@ -306,7 +296,6 @@ def test_multiply_array(ctx_factory):
     assert (a * a == a_squared).all()
 
 
-@pytools.test.mark_test.opencl
 def test_addition_array(ctx_factory):
     """Test the addition of two arrays."""
 
@@ -320,7 +309,6 @@ def test_addition_array(ctx_factory):
     assert (a + a == a_added).all()
 
 
-@pytools.test.mark_test.opencl
 def test_addition_scalar(ctx_factory):
     """Test the addition of an array and a scalar."""
 
@@ -334,7 +322,6 @@ def test_addition_scalar(ctx_factory):
     assert (7 + a == a_added).all()
 
 
-@pytools.test.mark_test.opencl
 def test_substract_array(ctx_factory):
     """Test the substraction of two arrays."""
     #test data
@@ -355,7 +342,6 @@ def test_substract_array(ctx_factory):
     assert (b - a == result).all()
 
 
-@pytools.test.mark_test.opencl
 def test_substract_scalar(ctx_factory):
     """Test the substraction of an array and a scalar."""
 
@@ -375,7 +361,6 @@ def test_substract_scalar(ctx_factory):
     assert (7 - a == result).all()
 
 
-@pytools.test.mark_test.opencl
 def test_divide_scalar(ctx_factory):
     """Test the division of an array and a scalar."""
 
@@ -392,7 +377,6 @@ def test_divide_scalar(ctx_factory):
     assert (np.abs(2 / a - result) < 1e-5).all()
 
 
-@pytools.test.mark_test.opencl
 def test_divide_array(ctx_factory):
     """Test the division of an array and a scalar. """
 
@@ -417,7 +401,6 @@ def test_divide_array(ctx_factory):
 
 # {{{ RNG
 
-@pytools.test.mark_test.opencl
 def test_random(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -461,7 +444,6 @@ def test_random(ctx_factory):
 
 # {{{ misc
 
-@pytools.test.mark_test.opencl
 def test_numpy_integer_shape(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -470,7 +452,6 @@ def test_numpy_integer_shape(ctx_factory):
     cl_array.empty(queue, (np.int32(17), np.int32(17)), np.float32)
 
 
-@pytools.test.mark_test.opencl
 def test_len(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -480,7 +461,6 @@ def test_len(ctx_factory):
     assert len(a_cpu) == 10
 
 
-@pytools.test.mark_test.opencl
 def test_stride_preservation(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -493,7 +473,6 @@ def test_stride_preservation(ctx_factory):
     assert np.allclose(AT_GPU.get(), AT)
 
 
-@pytools.test.mark_test.opencl
 def test_nan_arithmetic(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -519,7 +498,6 @@ def test_nan_arithmetic(ctx_factory):
     assert (np.isnan(ab) == np.isnan(ab_gpu)).all()
 
 
-@pytools.test.mark_test.opencl
 def test_mem_pool_with_arrays(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -532,7 +510,6 @@ def test_mem_pool_with_arrays(ctx_factory):
     assert b_dev.allocator is mem_pool
 
 
-@pytools.test.mark_test.opencl
 def test_view(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -552,10 +529,26 @@ def test_view(ctx_factory):
     view = a_dev.view(np.int16)
     assert view.shape == (8, 32) and view.dtype == np.int16
 
+
+def test_diff(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    from pyopencl.clrandom import rand as clrand
+
+    l = 20000
+    a_dev = clrand(queue, (l,), dtype=np.float32)
+    a = a_dev.get()
+
+    err = la.norm(
+            (cl.array.diff(a_dev).get() - np.diff(a)))
+    assert err < 1e-4
+
 # }}}
 
 
-@pytools.test.mark_test.opencl
+# {{{ slices, concatenation
+
 def test_slice(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -597,7 +590,6 @@ def test_slice(ctx_factory):
         assert la.norm(a_gpu.get() - a) == 0
 
 
-@pytools.test.mark_test.opencl
 def test_concatenate(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -616,6 +608,109 @@ def test_concatenate(ctx_factory):
 
     assert la.norm(cat - cat_dev.get()) == 0
 
+# }}}
+
+
+# {{{ conditionals, any, all
+
+def test_comparisons(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    from pyopencl.clrandom import rand as clrand
+
+    l = 20000
+    a_dev = clrand(queue, (l,), dtype=np.float32)
+    b_dev = clrand(queue, (l,), dtype=np.float32)
+
+    a = a_dev.get()
+    b = b_dev.get()
+
+    import operator as o
+    for op in [o.eq, o.ne, o.le, o.lt, o.ge, o.gt]:
+        res_dev = op(a_dev, b_dev)
+        res = op(a, b)
+
+        assert (res_dev.get() == res).all()
+
+        res_dev = op(a_dev, 0)
+        res = op(a, 0)
+
+        assert (res_dev.get() == res).all()
+
+        res_dev = op(0, b_dev)
+        res = op(0, b)
+
+        assert (res_dev.get() == res).all()
+
+
+def test_any_all(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    l = 20000
+    a_dev = cl_array.zeros(queue, (l,), dtype=np.int8)
+
+    assert not a_dev.all()
+    assert not a_dev.any()
+
+    a_dev[15213] = 1
+
+    assert not a_dev.all()
+    assert a_dev.any()
+
+    a_dev.fill(1)
+
+    assert a_dev.all()
+    assert a_dev.any()
+
+# }}}
+
+
+def test_map_to_host(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    if context.devices[0].type & cl.device_type.GPU:
+        mf = cl.mem_flags
+        allocator = cl_tools.DeferredAllocator(
+                context, mf.READ_WRITE | mf.ALLOC_HOST_PTR)
+    else:
+        allocator = None
+
+    a_dev = cl_array.zeros(queue, (5, 6, 7,), dtype=np.float32, allocator=allocator)
+    a_dev[3, 2, 1] = 10
+    a_host = a_dev.map_to_host()
+    a_host[1, 2, 3] = 10
+
+    a_host_saved = a_host.copy()
+    a_host.base.release(queue)
+
+    a_dev.finish()
+
+    print("DEV[HOST_WRITE]", a_dev.get()[1, 2, 3])
+    print("HOST[DEV_WRITE]", a_host_saved[3, 2, 1])
+
+    assert (a_host_saved == a_dev.get()).all()
+
+
+def test_view_and_strides(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    from pyopencl.clrandom import rand as clrand
+
+    X = clrand(queue, (5, 10), dtype=np.float32)
+    Y = X[:3, :5]
+    y = Y.view()
+
+    assert y.shape == Y.shape
+    assert y.strides == Y.strides
+
+    import pytest
+    with pytest.raises(AssertionError):
+        assert (y.get() == X.get()[:3, :5]).all()
+
 
 if __name__ == "__main__":
     # make sure that import failures get reported, instead of skipping the
